%--------------------------------------------------------
% >>> rmseed.m <<< utility for reading binary MSEED files
%--------------------------------------------------------
%
% Calling convention
% [info,x] = rmseed('in_file');
%
% Anytime keep care about the output variable info.err_flg
% If info.err_flg <> 0, do not believe other output values
% are correct.
%
% Co.B.R. Dec.2009
% Comments please send to b.ruzek@ig.cas.cz
%
% On input: in_file ... name of the MSeed file. Till now, it should be
%           demultiplexed file only and coded using Steim1 compression
%           scheme. Little/Big endian is corrected for.
% On output: info.err_flg ... bit encoded error (normally should be 0)
%            info.stat    ... code (name) of the station
%            info.stream  ... channel ID following FDSN specifications
%            info.FS      ... sampling frequency in [Hz]
%            info.Ts      ... time of the first sample in seconds
%                             (including fractions of second) since
%                             January 1970, 00:00:00.000
%            info.date    ... time of the first sample as ASCII string
%            x            ... output signal
% NOTE1: Possible gaps in the signal are filled with zeros, possible
% overlays are solved by keeping only the last version of the signal.
% NOTE2: Please select correct coding according your CPU platform
% (variable 'endian' just on the first line in rmseed())

function [info,x] = rmseed(in_file)


endian    = 0; % Little endian coding (PC,HP,Intel_CPU_family,...)
%endian    = 1; % Big endian coding (Sun, Macintosh,...)
LEN_FRAME =  64;
MAXCHAR   = 2^7;
RNGCHAR   = 2^8;
MAXSHORT  = 2^15;
RNGSHORT  = 2^16;
MAXLONG   = 2^31;
RNGLONG   = 2^32;
RNG3      = 2^3;
RNG4      = 2^4;
RNG5      = 2^5;
RNG6      = 2^6;
RNG9      = 2^9;
RNG10     = 2^10;
RNG14     = 2^14;
RNG15     = 2^15;
RNG29     = 2^29;
RNG30     = 2^30;
MS_file_info = [];
buffer    = [];
x         = [];
info      = [];
if(Fill_info(in_file) == 0)
    if(Check_hdr() == 0)
        n_block = MS_file_info.n_block;
        FS      = MS_file_info.hdr(1,3);
        T       = MS_file_info.hdr(n_block,7) - MS_file_info.hdr(1,6);
        n_samp  = T*FS;
        x       = zeros(n_samp,1);
        for i=1:MS_file_info.n_block
            data = Decode_Data(i);
            ndat = size(data,1);
            is   = int32((MS_file_info.hdr(i,6)-MS_file_info.hdr(1,6))*FS) + 1;
            ie   = is + ndat - 1;
            x(is:ie) = data;
        end
    end
    info.stat   = MS_file_info.stat;
    info.stream = MS_file_info.stream;
    info.FS     = MS_file_info.hdr(1,3);
    info.Ts     = MS_file_info.hdr(1,6);
    info.date   = MS_file_info.date;
end
info.err_flg = MS_file_info.err_flg;
return;

%------------------------------
    function secs = gettime(dt)
        dt(2) = dt(2) - 2;
        if(dt(2) <= 0)
            dt(2) = dt(2) + 12;
            dt(1) = dt(1) - 1;
        end
        secs = ((((int32(dt(1)/4)-int32(dt(1)/100)+int32(dt(1)/400)+int32(367*dt(2)/12)+dt(3))+dt(1)*365-719499)*24 +dt(4))*60+dt(5))*60+dt(6);
    end

%------------------------------------
    function err = Fill_info(in_file)
        err = 0;
        MS_file_info = [];
        MS_file_info.err_flg  = 0;
        MS_file_info.in_file  = in_file;
        MS_file_info.n_block  = 0;
        MS_file_info.rec_len  = 0;
        MS_file_info.date     = '';
        fp_i = fopen(MS_file_info.in_file, 'rb');
        if(fp_i < 0)
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^8);
            err = 2^8;
            return;
        end
        buffer = fread(fp_i);
        fclose(fp_i);
        MS_file_info.rec_len = 2^double(buffer(55));
        MS_file_info.n_block = size(buffer,1)/MS_file_info.rec_len;
        MS_file_info.hdr     = zeros(MS_file_info.n_block,7);
        for ii=1:MS_file_info.n_block
            ihs = (ii-1)*MS_file_info.rec_len+1;
            ihe = ihs + LEN_FRAME;
            [stat,stream,swap_flg,hdr] = Decode_Header(buffer(ihs:ihe));
            MS_file_info.hdr(ii,:) = hdr;
        end
        MS_file_info.swap_flg = swap_flg;
        MS_file_info.stat     = stat;
        MS_file_info.stream   = stream;
        return;
    end


%-------------------------------------------------------------
    function [stat,stream,swap_flg,hdr] = Decode_Header(bufhdr)
        hdr    = zeros(1,7);
        stat   = '';
        stream = '';
        if(char(bufhdr(7)) ~= 'D' && char(bufhdr(7)) ~= 'R')
            return;
        end
        Block_info.cur_blk = sscanf(char(bufhdr(1:6)), '%d');
        stat = sscanf(char(bufhdr(9:13)), '%s');
        stream = char(bufhdr(16:18))';
        Block_info.stream = bufhdr(16)+256*(bufhdr(17) + 256*bufhdr(18));
        encod_fmt  = int32(bufhdr(53));
        word_order = int32(bufhdr(54));
        if((word_order == 1 && endian == 0)||(word_order == 0 && endian == 1))
            swap_flg = 1;
        else
            swap_flg = 0;
        end
        if(swap_flg)
            if(encod_fmt == 10)
                Decode_Data = @Decode_Data_S1;
            else
                Decode_Data = @Decode_Data_S2;
            end
            year    = 256*bufhdr(21) + bufhdr(22);
            jday    = 256*bufhdr(23) + bufhdr(24);
            hour    = bufhdr(25);
            tmin    = bufhdr(26);
            sec     = bufhdr(27);
            curms   = 256*bufhdr(29) + bufhdr(30);
            %            corrms  = 256*double(bufhdr(41)) + double(bufhdr(42));
            fs      = 256*bufhdr(33) + bufhdr(34);
            ms      = 256*bufhdr(35) + bufhdr(36);
            Block_info.samples  = 256*bufhdr(31) + bufhdr(32);
        else
            if(encod_fmt == 10)
                Decode_Data = @Decode_Data_s1;
            else
                Decode_Data = @Decode_Data_S2;
            end
            year    = 256*bufhdr(22) + bufhdr(21);
            jday    = 256*bufhdr(24) + bufhdr(23);
            hour    = bufhdr(25);
            tmin    = bufhdr(26);
            sec     = bufhdr(27);
            curms   = 256*bufhdr(30) + bufhdr(29);
            %            corrms  = 256*double(bufhdr(42)) + double(bufhdr(41));
            fs      = 256*bufhdr(34) + bufhdr(33);
            ms      = 256*bufhdr(36) + bufhdr(35);
            Block_info.samples  = 256*bufhdr(32) + bufhdr(31);
        end
        if(fs > MAXSHORT)
            fs = fs - RNGSHORT;
        end
        if(ms > MAXSHORT)
            ms = ms - RNGSHORT;
        end
        if(fs > 0)
            if(ms > 0)
                Block_info.Fs = fs*ms;
            else
                Block_info.Fs = -fs/ms;
            end
        else
            if(ms > 0)
                Block_info.Fs = -ms/fs;
            else
                Block_info.Fs = 1/(ms*fs);
            end
        end
        tc = [year 01 01 hour tmin sec+curms/10000];
        Block_info.blk_len  = (Block_info.samples-1)/Block_info.Fs;
        Block_info.Blk_S  = gettime(tc) + (jday-1)*86400;        % start-time in seconds
        Block_info.Blk_E  = Block_info.Blk_S + Block_info.blk_len;
        hdr(1) = Block_info.cur_blk;
        hdr(2) = Block_info.samples;
        hdr(3) = Block_info.Fs;
        hdr(4) = Block_info.stream;
        hdr(5) = Block_info.blk_len;
        hdr(6) = Block_info.Blk_S;
        hdr(7) = Block_info.Blk_E;
        if(length(MS_file_info.date) == 0)
            MS_file_info.date = sprintf('%4.4d_%3.3d_%2.2d:%2.2d:%6.3f', year, jday, hour, tmin, sec+curms/10000);
        end
        return
    end


%-----------------------------
    function err = Check_hdr()
        if(min(MS_file_info.hdr(:,3)) ~= max(MS_file_info.hdr(:,3)))
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^0);
        end
        if(min(MS_file_info.hdr(:,4)) ~= max(MS_file_info.hdr(:,4)))
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^1);
        end
        if(min(diff(MS_file_info.hdr(:,6))) < 0)
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^2);
        end
        if(min(diff(MS_file_info.hdr(:,7))) < 0)
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^3);
        end
        err = MS_file_info.err_flg;
        return
    end


%---------------------------------------
    function [data] = Decode_Data_S1(i_b)
        % Steim 1 compression scheme + swapping enabled
        data = [];
        ii = 0;
        mask = zeros(1,15);
        i_s = (i_b-1)*MS_file_info.rec_len+1;
        i_e = i_s + MS_file_info.rec_len - 1;
        data_off = i_s + double(256*buffer(i_s+44)+buffer(i_s+45));
        sample_0 = ((buffer(data_off+4)*256+buffer(data_off+5))*256+buffer(data_off+6))*256+buffer(data_off+7);
        sample_n = ((buffer(data_off+8)*256+buffer(data_off+9))*256+buffer(data_off+10))*256+buffer(data_off+11);
        if(sample_0 >= MAXLONG)
            sample_0 = double(sample_0) - RNGLONG;
        end
        if(sample_n >= MAXLONG)
            sample_n = double(sample_n) - RNGLONG;
        end
        for i_f=data_off:LEN_FRAME:i_e
            EC = uint32(((buffer(i_f)*256+buffer(i_f+1))*256+buffer(i_f+2))*256+buffer(i_f+3));
            for nib=1:15
                mask(nib) = bitand(EC,3);
                EC = uint32(floor(double(EC)/4));
            end
            i_d = i_f+4;
            for nib=15:-1:1
                switch mask(nib)
                    case 0
                        i_d=i_d+4;
                        continue;
                    case 1
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                    case 2
                        ii=ii+1;data(ii)=buffer(i_d)*256+buffer(i_d+1);i_d=i_d+2;
                        if(data(ii) >= MAXSHORT)
                            data(ii) = data(ii) - RNGSHORT;
                        end
                        ii=ii+1;data(ii)=buffer(i_d)*256+buffer(i_d+1);i_d=i_d+2;
                        if(data(ii) >= MAXSHORT)
                            data(ii) = data(ii) - RNGSHORT;
                        end
                    case 3
                        ii=ii+1;data(ii)=((buffer(i_d)*256+buffer(i_d+1))*256+buffer(i_d+2))*256+buffer(i_d+3);i_d=i_d+4;
                        if(data(ii) >= MAXLONG)
                            data(ii) = data(ii) - RNGLONG;
                        end
                end
            end
        end
        data(1) = sample_0;
        data = cumsum(data)';
        sample_d = data(size(data,1));
        if(sample_d ~= sample_n)
            MS_file_info.err_flg = bitor(MS_file_info,err_flg,2^5);
        end
        if(size(data,1) ~= MS_file_info.hdr(i_b,2))
            MS_file_info.err_flg = bitor(MS_file_info,err_flg,2^6);
        end
        return;
    end


%---------------------------------------
    function [data] = Decode_Data_s1(i_b)
        % Steim 1 compression scheme + swapping disabled
        data = [];
        ii = 0;
        mask = zeros(1,15);
        i_s = (i_b-1)*MS_file_info.rec_len+1;
        i_e = i_s + MS_file_info.rec_len - 1;
        data_off = i_s + double(256*buffer(i_s+45)+buffer(i_s+44));
        sample_0 = ((buffer(data_off+7)*256+buffer(data_off+6))*256+buffer(data_off+5))*256+buffer(data_off+4);
        sample_n = ((buffer(data_off+11)*256+buffer(data_off+10))*256+buffer(data_off+9))*256+buffer(data_off+8);
        if(sample_0 >= MAXLONG)
            sample_0 = double(sample_0) - RNGLONG;
        end
        if(sample_n >= MAXLONG)
            sample_n = double(sample_n) - RNGLONG;
        end
        for i_f=data_off:LEN_FRAME:i_e
            EC = uint32(((buffer(i_f+3)*256+buffer(i_f+2))*256+buffer(i_f+1))*256+buffer(i_f));
            for nib=1:15
                mask(nib) = bitand(EC,3);
                EC = uint32(floor(double(EC)/4));
            end
            i_d = i_f+4;
            for nib=15:-1:1
                switch mask(nib)
                    case 0
                        i_d=i_d+4;
                        continue;
                    case 1
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                    case 2
                        ii=ii+1;data(ii)=buffer(i_d+1)*256+buffer(i_d);i_d=i_d+2;
                        if(data(ii) >= MAXSHORT)
                            data(ii) = data(ii) - RNGSHORT;
                        end
                        ii=ii+1;data(ii)=buffer(i_d+1)*256+buffer(i_d);i_d=i_d+2;
                        if(data(ii) >= MAXSHORT)
                            data(ii) = data(ii) - RNGSHORT;
                        end
                    case 3
                        ii=ii+1;data(ii)=((buffer(i_d+3)*256+buffer(i_d+2))*256+buffer(i_d+1))*256+buffer(i_d);i_d=i_d+4;
                        if(data(ii) >= MAXLONG)
                            data(ii) = data(ii) - RNGLONG;
                        end
                end
            end
        end
        data(1) = sample_0;
        data = cumsum(data)';
        sample_d = data(size(data,1));
        if(sample_d ~= sample_n)
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^5);
        end
        if(size(data,1) ~= MS_file_info.hdr(i_b,2))
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^6);
        end
        return;
    end


%---------------------------------------
    function [data] = Decode_Data_S2(i_b)
        % Steim 2 compression scheme + swap check
        data = [];
        ii = 0;
        mask = zeros(1,15);
        i_s = (i_b-1)*MS_file_info.rec_len+1;
        i_e = i_s + MS_file_info.rec_len - 1;
        if(MS_file_info.swap_flg)
            data_off = i_s + double(256*buffer(i_s+44)+buffer(i_s+45));
            sample_0 = ((buffer(data_off+4)*256+buffer(data_off+5))*256+buffer(data_off+6))*256+buffer(data_off+7);
            sample_n = ((buffer(data_off+8)*256+buffer(data_off+9))*256+buffer(data_off+10))*256+buffer(data_off+11);
        else
            data_off = i_s + double(256*buffer(i_s+45)+buffer(i_s+44));
            sample_0 = ((buffer(data_off+7)*256+buffer(data_off+6))*256+buffer(data_off+5))*256+buffer(data_off+4);
            sample_n = ((buffer(data_off+11)*256+buffer(data_off+10))*256+buffer(data_off+9))*256+buffer(data_off+8);
        end
        if(sample_0 >= MAXLONG)
            sample_0 = double(sample_0) - RNGLONG;
        end
        if(sample_n >= MAXLONG)
            sample_n = double(sample_n) - RNGLONG;
        end
        for i_f=data_off:LEN_FRAME:i_e
            if(MS_file_info.swap_flg)
                EC = uint32(((buffer(i_f)*256+buffer(i_f+1))*256+buffer(i_f+2))*256+buffer(i_f+3));
            else
                EC = uint32(((buffer(i_f+3)*256+buffer(i_f+2))*256+buffer(i_f+1))*256+buffer(i_f));
            end
            for nib=1:15
                mask(nib) = bitand(EC,3);
                EC = uint32(floor(double(EC)/4));
            end
            i_d = i_f+4;
            for nib=15:-1:1
                switch mask(nib)
                    case 0
                        i_d=i_d+4;
                        continue;
                    case 1
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                        ii=ii+1;data(ii)=buffer(i_d);i_d=i_d+1;
                        if(data(ii) >= MAXCHAR)
                            data(ii) = data(ii) - RNGCHAR;
                        end
                    case 2
                        if(MS_file_info.swap_flg == 0)
                            aux = buffer(i_d); buffer(i_d) = buffer(i_d+3); buffer(i_d+3) = aux;
                            aux = buffer(i_d+1); buffer(i_d+1) = buffer(i_d+2); buffer(i_d+2) = aux;
                        end
                        dnib = floor(buffer(i_d)/64);
                        switch dnib
                            case 1
                                ii = ii+1;data(ii)=((double(bitand(buffer(i_d),63))*256+double(buffer(i_d+1)))*256+double(buffer(i_d+2)))*256+double(buffer(i_d+3));
                                if(data(ii) >= RNG29), data(ii) = data(ii)-RNG30; end
                                i_d = i_d + 4;
                            case 2
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d),63))*512+double(buffer(i_d+1))*2+double(bitand(buffer(i_d+2),128))/128;
                                if(data(ii) >= RNG14), data(ii) = data(ii)-RNG15; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),127))*256+double(buffer(i_d+3));
                                if(data(ii) >= RNG14), data(ii) = data(ii)-RNG15; end
                                i_d = i_d + 4;
                            case 3
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d),63))*16+double(bitand(buffer(i_d+1),240))/16;
                                if(data(ii) >= RNG9), data(ii) = data(ii)-RNG10; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),15))*64+double(bitand(buffer(i_d+2),252))/4;
                                if(data(ii) >= RNG9), data(ii) = data(ii)-RNG10; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),3))*256+double(buffer(i_d+3));
                                if(data(ii) >= RNG9), data(ii) = data(ii)-RNG10; end
                                i_d = i_d + 4;
                        end
                    case 3
                        if(MS_file_info.swap_flg == 0)
                            aux = buffer(i_d); buffer(i_d) = buffer(i_d+3); buffer(i_d+3) = aux;
                            aux = buffer(i_d+1); buffer(i_d+1) = buffer(i_d+2); buffer(i_d+2) = aux;
                        end
                        dnib = floor(buffer(i_d)/64);
                        switch dnib
                            case 0
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d),63));
                                if(data(ii) >= RNG5), data(ii) = data(ii)-RNG6; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),252))/4;
                                if(data(ii) >= RNG5), data(ii) = data(ii)-RNG6; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),3))*16+double(bitand(buffer(i_d+2),240))/16;
                                if(data(ii) >= RNG5), data(ii) = data(ii)-RNG6; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),15))*4+double(bitand(buffer(i_d+3),192))/64;
                                if(data(ii) >= RNG5), data(ii) = data(ii)-RNG6; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+3),63));
                                if(data(ii) >= RNG5), data(ii) = data(ii)-RNG6; end
                                i_d = i_d + 4;
                            case 1
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d),62))/2;
                                if(data(ii) >= RNG4), data(ii) = data(ii)-RNG5; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d),1))*16+double(bitand(buffer(i_d+1),240))/16;
                                if(data(ii) >= RNG4), data(ii) = data(ii)-RNG5; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),15))*2+double(bitand(buffer(i_d+2),128))/128;
                                if(data(ii) >= RNG4), data(ii) = data(ii)-RNG5; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),124))/4;
                                if(data(ii) >= RNG4), data(ii) = data(ii)-RNG5; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),3))*8+double(bitand(buffer(i_d+3),224))/32;
                                if(data(ii) >= RNG4), data(ii) = data(ii)-RNG5; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+3),31));
                                if(data(ii) >= RNG4), data(ii) = data(ii)-RNG5; end
                                i_d = i_d + 4;
                            case 2
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d),15));
                                if(data(ii) >= RNG3), data(ii) = data(ii)-RNG4; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),240))/16;
                                if(data(ii) >= RNG3), data(ii) = data(ii)-RNG4; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),15));
                                if(data(ii) >= RNG3), data(ii) = data(ii)-RNG4; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),240))/16;
                                if(data(ii) >= RNG3), data(ii) = data(ii)-RNG4; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),15));
                                if(data(ii) >= RNG3), data(ii) = data(ii)-RNG4; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+3),240))/16;
                                if(data(ii) >= RNG3), data(ii) = data(ii)-RNG4; end
                                ii = ii+1;data(ii)=double(bitand(buffer(i_d+3),15));
                                if(data(ii) >= RNG3), data(ii) = data(ii)-RNG4; end
                                i_d = i_d + 4;
                        end
                end
            end
        end
        data(1) = sample_0;
        data = cumsum(data)';
        sample_d = data(size(data,1));
        if(sample_d ~= sample_n)
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^5);
        end
        if(size(data,1) ~= MS_file_info.hdr(i_b,2))
            MS_file_info.err_flg = bitor(MS_file_info.err_flg,2^6);
        end
        return;
    end
end
