// >>> RMSeed.sce <<< utility for reading binary MSEED files
//
// Co.B.R. Dec.2009
//

clear

BLK_LEN   = 4096;
MAX_FRAME =   64;
LEN_FRAME =   64;
MAXSHORT  = 2^15;
RNGSHORT  = 2^16;
MAXLONG   = 2^31;
RNGLONG   = 2^32;
RNG3      = 2^3;
RNG4      = 2^4;
RNG5      = 2^5;
RNG6      = 2^6;
RNG9      = 2^9;
RNG10     = 2^10;
RNG14     = 2^14;
RNG15     = 2^15;
RNG29     = 2^29;
RNG30     = 2^30;
MS_file_info = [];

//------------------------
function secs = gettime(dt)
  dt(2) = dt(2) - 2;
  if(dt(2) <= 0)
    dt(2) = dt(2) + 12;
    dt(1) = dt(1) - 1;
  end
  secs = ((((int(dt(1)/4)-int(dt(1)/100)+int(dt(1)/400)+int(367*dt(2)/12)+dt(3))+dt(1)*365-719499)*24 +dt(4))*60+dt(5))*60+dt(6);
endfunction

//---------------------------
function [info,x] = RMSeed(in_file)
  global MS_file_info;
  x    = [];
  info = [];
  if(Fill_info(in_file) == 0)
    if(Check_hdr() == 0)
      n_block = MS_file_info.n_block;
      FS      = MS_file_info.hdr(1,3);
      T       = MS_file_info.hdr(n_block,7) - MS_file_info.hdr(1,6);
      n_samp  = T*FS;
      x       = zeros(n_samp,1);
      for i=1:MS_file_info.n_block
        if(MS_file_info.encod_fmt == 10)
          data = Decode_DataS1(i);
        end
        if(MS_file_info.encod_fmt == 11)
          data = Decode_DataS2(i);
        end
        ndat = size(data,1);
        is   = int((MS_file_info.hdr(i,6)-MS_file_info.hdr(1,6))*FS) + 1;
        ie   = is + ndat - 1;
        x(is:ie) = data;
      end
    end
    info.stat   = MS_file_info.stat;
    info.stream = MS_file_info.stream;
    info.FS     = MS_file_info.hdr(1,3);
    info.Ts     = MS_file_info.hdr(1,6);
    info.encod  = MS_file_info.encod_fmt;
    info.endian = MS_file_info.word_order;
  end
  info.err_flg = MS_file_info.err_flg;
  clear MS_file_info;
  clear data;
  return;
endfunction

//------------------------
function err = Fill_info(in_file)
  global MS_file_info;
  global buffer;
  err = 0;
  MS_file_info = [];
  MS_file_info.err_flg  = 0;
  MS_file_info.in_file  = in_file;
  MS_file_info.n_block  = 0;
  MS_file_info.rec_len  = 0;
  [f_info,f_err] = fileinfo(MS_file_info.in_file);
  if(f_err ~= 0)
    err = 2^8;
    return;
  end
  fp_i = mopen(MS_file_info.in_file, 'rb');
  buffer = mgeti(f_info(1), 'uc', fp_i);
  mclose(fp_i);
  MS_file_info.rec_len = 2^double(buffer(55));
  MS_file_info.n_block = f_info(1)/MS_file_info.rec_len;
  MS_file_info.hdr     = zeros(MS_file_info.n_block,7);
  for i=1:MS_file_info.n_block
    ihs = (i-1)*MS_file_info.rec_len+1;
    ihe = ihs + LEN_FRAME;
    [stat,stream,swap_flg,hdr] = Decode_Header(buffer(ihs:ihe));
    MS_file_info.hdr(i,:) = hdr;
  end
  MS_file_info.swap_flg = swap_flg;
  MS_file_info.stat     = stat;
  MS_file_info.stream   = stream;
  return;
endfunction

//-----------------------------------------------
function [stat,stream,swap_flg,hdr] = Decode_Header(buffer)
  global MS_file_info;
  hdr    = zeros(1,7);
  stat   = '';
  stream = '';
  if(char(buffer(7)) ~= 'D' & char(buffer(7)) ~= 'R')
    return;
  end
  Block_info.cur_blk = msscanf(char(buffer(1:6)), '%d');
  stat = msscanf(char(buffer(9:13)), '%s');
  stream = char(buffer(16:18));
  Block_info.stream = double(buffer(16))+256*(double(buffer(17)) + 256*double(buffer(18)));
  encod_fmt  = int(buffer(53));
  word_order = int(buffer(54));
  if(word_order == 1)
    swap_flg = 1;
  else
    swap_flg = 0;
  end
  MS_file_info.word_order = double(word_order);
  MS_file_info.encod_fmt  = double(encod_fmt);
  if(swap_flg)
    year    = 256*double(buffer(21)) + double(buffer(22));
    jday    = 256*double(buffer(23)) + double(buffer(24));
    hour    = double(buffer(25));
    tmin    = double(buffer(26));
    sec     = double(buffer(27));
    curms   = 256*double(buffer(29)) + double(buffer(30));
    corrms  = 256*double(buffer(41)) + double(buffer(42));
    fs      = 256*iconvert(buffer(33),14) + iconvert(buffer(34),14);
    ms      = 256*iconvert(buffer(35),14) + iconvert(buffer(36),14);
    Block_info.samples  = 256*double(buffer(31)) + double(buffer(32));
  else
    year    = 256*double(buffer(22)) + double(buffer(21));
    jday    = 256*double(buffer(24)) + double(buffer(23));
    hour    = double(buffer(25));
    tmin    = double(buffer(26));
    sec     = double(buffer(27));
    curms   = 256*double(buffer(30)) + double(buffer(29));
    corrms  = 256*double(buffer(42)) + double(buffer(41));
    fs      = 256*iconvert(buffer(34),14) + iconvert(buffer(33),14);
    ms      = 256*iconvert(buffer(36),14) + iconvert(buffer(35),14);
    Block_info.samples  = 256*double(buffer(32)) + double(buffer(31));
  end
  if(fs > MAXSHORT) fs = fs - RNGSHORT; end
  if(ms > MAXSHORT) ms = ms - RNGSHORT; end
  if(fs > 0)
    if(ms > 0)
	    Block_info.Fs = fs*ms;
	   else
	    Block_info.Fs = -fs/ms;
    end
  else
    if(ms > 0)
	    Block_info.Fs = -ms/fs;
	   else
      Block_info.Fs = 1/(ms*fs);
    end
  end
  tc = [year 01 01 hour tmin sec+curms/10000];
  Block_info.blk_len  = (Block_info.samples-1)/Block_info.Fs;
  Block_info.Blk_S  = gettime(tc) + (jday-1)*86400;        // start-time in seconds
  Block_info.Blk_E  = Block_info.Blk_S + Block_info.blk_len;
  hdr(1) = Block_info.cur_blk;
  hdr(2) = Block_info.samples;
  hdr(3) = Block_info.Fs;
  hdr(4) = Block_info.stream;
  hdr(5) = Block_info.blk_len;
  hdr(6) = Block_info.Blk_S;
  hdr(7) = Block_info.Blk_E;
  return
endfunction


//------------------
function err = Check_hdr()
  global MS_file_info;
  if(min(MS_file_info.hdr(:,3)) ~= max(MS_file_info.hdr(:,3)))
    MS_file_info.err = bitor(MS_file_info.err,2^0);
  end
  if(min(MS_file_info.hdr(:,4)) ~= max(MS_file_info.hdr(:,4)))
    MS_file_info.err = bitor(MS_file_info.err,2^1);
  end
  if(min(diff(MS_file_info.hdr(:,6))) < 0)
    MS_file_info.err = bitor(MS_file_info.err,2^2);
  end
  if(min(diff(MS_file_info.hdr(:,7))) < 0)
    MS_file_info.err = bitor(MS_file_info.err,RNG3);
  end
  err = MS_file_info.err_flg;
  return
endfunction


//------------------------
function nn = bitor(na, nb)
  nn = 0;
  ii = 1;
  while(na ~= 0 | nb ~= 0)
    nam = modulo(na,2);
    nbm = modulo(nb,2);
    if(nam == 1 | nbm == 1)
      nn = nn + ii;
    end
    ii = ii*2;
    na = int(na/2);
    nb = int(nb/2);
  end
  return;
endfunction


//------------------------
function nn = bitand(na, nb)
  nn = 0;
  ii = 1;
  while(na ~= 0 & nb ~= 0)
    nam = modulo(na,2);
    nbm = modulo(nb,2);
    if(nam == 1 & nbm == 1)
      nn = nn + ii;
    end
    ii = ii*2;
    na = int(na/2);
    nb = int(nb/2);
  end
  return;
endfunction


//---------------------------------
function [data] = Decode_DataS2(i_b)
  global MS_file_info;
  global buffer;
  data = [];
  ii = 0;
  mask = zeros(1,15);
  i_s = (i_b-1)*MS_file_info.rec_len+1;
  i_e = i_s + MS_file_info.rec_len - 1;
  if(MS_file_info.swap_flg)
    data_off = i_s + double(256*buffer(i_s+44)+buffer(i_s+45));
    sample_0 = ((iconvert(buffer(data_off+4),14)*256+iconvert(buffer(data_off+5),14))*256+iconvert(buffer(data_off+6),14))*256+iconvert(buffer(data_off+7),14);
    sample_n = ((iconvert(buffer(data_off+8),14)*256+iconvert(buffer(data_off+9),14))*256+iconvert(buffer(data_off+10),14))*256+iconvert(buffer(data_off+11),14);
  else
    data_off = i_s + double(256*buffer(i_s+45)+buffer(i_s+44));
    sample_0 = ((iconvert(buffer(data_off+7),14)*256+iconvert(buffer(data_off+6),14))*256+iconvert(buffer(data_off+5),14))*256+iconvert(buffer(data_off+4),14);
    sample_n = ((iconvert(buffer(data_off+11),14)*256+iconvert(buffer(data_off+10),14))*256+iconvert(buffer(data_off+9),14))*256+iconvert(buffer(data_off+8),14);
  end
  if(sample_0 >= MAXLONG) sample_0 = double(sample_0) - RNGLONG; end
  if(sample_n >= MAXLONG) sample_n = double(sample_n) - RNGLONG; end
  for i_f=data_off:LEN_FRAME:i_e
    if(MS_file_info.swap_flg)
      EC = ((iconvert(buffer(i_f),14)*256+iconvert(buffer(i_f+1),14))*256+iconvert(buffer(i_f+2),14))*256+iconvert(buffer(i_f+3),14);
    else
      EC = ((iconvert(buffer(i_f+3),14)*256+iconvert(buffer(i_f+2),14))*256+iconvert(buffer(i_f+1),14))*256+iconvert(buffer(i_f),14);
    end
//    mprintf('%d %d %d\n',sample_0, sample_n, EC);
    for nib=1:15
      mask(nib) = modulo(EC,4);
      EC = int(EC/4);
//     mprintf('%d ',mask(nib));
    end
    i_d = i_f+4;
    for nib=15:-1:1
      select mask(nib)
      case 0
        i_d=i_d+4;
        continue;
      case 1
        ii=ii+1;data(ii)=double(iconvert(buffer(i_d),1));i_d=i_d+1;
        ii=ii+1;data(ii)=double(iconvert(buffer(i_d),1));i_d=i_d+1;
        ii=ii+1;data(ii)=double(iconvert(buffer(i_d),1));i_d=i_d+1;
        ii=ii+1;data(ii)=double(iconvert(buffer(i_d),1));i_d=i_d+1;
      case 2
        if(MS_file_info.swap_flg == 0)
          aux = buffer(i_d); buffer(i_d) = buffer(i_d+3); buffer(i_d+3) = aux;
          aux = buffer(i_d+1); buffer(i_d+1) = buffer(i_d+2); buffer(i_d+2) = aux;
          dnib = int(iconvert(buffer(i_d),11)/64);
        end
        dnib = int(iconvert(buffer(i_d),11)/64);
        select dnib
          case 1
            ii = ii+1;data(ii)=((double(bitand(buffer(i_d),63))*256+double(buffer(i_d+1)))*256+double(buffer(i_d+2)))*256+double(buffer(i_d+3));
            if(data(ii) >= RNG29) data(ii) = data(ii)-RNG30; end
            i_d = i_d + 4;
          case 2
            ii = ii+1;data(ii)=double(bitand(buffer(i_d),63))*512+double(buffer(i_d+1))*2+double(bitand(buffer(i_d+2),128))/128;
            if(data(ii) >= RNG14) data(ii) = data(ii)-RNG15; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),127))*256+double(buffer(i_d+3));
            if(data(ii) >= RNG14) data(ii) = data(ii)-RNG15; end
            i_d = i_d + 4;
          case 3
            ii = ii+1;data(ii)=double(bitand(buffer(i_d),63))*16+double(bitand(buffer(i_d+1),240))/16;
            if(data(ii) >= RNG9) data(ii) = data(ii)-RNG10; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),15))*64+double(bitand(buffer(i_d+2),252))/4;
            if(data(ii) >= RNG9) data(ii) = data(ii)-RNG10; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),3))*256+double(buffer(i_d+3));
            if(data(ii) >= RNG9) data(ii) = data(ii)-RNG10; end
            i_d = i_d + 4;
          end
      case 3
        if(MS_file_info.swap_flg == 0)
          aux = buffer(i_d); buffer(i_d) = buffer(i_d+3); buffer(i_d+3) = aux;
          aux = buffer(i_d+1); buffer(i_d+1) = buffer(i_d+2); buffer(i_d+2) = aux;
          dnib = int(iconvert(buffer(i_d),11)/64);
        end
        dnib = int(iconvert(buffer(i_d),11)/64);
        select dnib
          case 0
            ii = ii+1;data(ii)=double(bitand(buffer(i_d),63));
            if(data(ii) >= RNG5) data(ii) = data(ii)-RNG6; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),252))/4;
            if(data(ii) >= RNG5) data(ii) = data(ii)-RNG6; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),3))*16+double(bitand(buffer(i_d+2),240))/16;
            if(data(ii) >= RNG5) data(ii) = data(ii)-RNG6; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),15))*4+double(bitand(buffer(i_d+3),192))/64;
            if(data(ii) >= RNG5) data(ii) = data(ii)-RNG6; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+3),63));
            if(data(ii) >= RNG5) data(ii) = data(ii)-RNG6; end
            i_d = i_d + 4;
          case 1
            ii = ii+1;data(ii)=double(bitand(buffer(i_d),62))/2;
            if(data(ii) >= RNG4) data(ii) = data(ii)-RNG5; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d),1))*16+double(bitand(buffer(i_d+1),240))/16;
            if(data(ii) >= RNG4) data(ii) = data(ii)-RNG5; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),15))*2+double(bitand(buffer(i_d+2),128))/128;
            if(data(ii) >= RNG4) data(ii) = data(ii)-RNG5; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),124))/4;
            if(data(ii) >= RNG4) data(ii) = data(ii)-RNG5; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),3))*8+double(bitand(buffer(i_d+3),224))/32;
            if(data(ii) >= RNG4) data(ii) = data(ii)-RNG5; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+3),31));
            if(data(ii) >= RNG4) data(ii) = data(ii)-RNG5; end
            i_d = i_d + 4;
          case 2
            ii = ii+1;data(ii)=double(bitand(buffer(i_d),15));
            if(data(ii) >= RNG3) data(ii) = data(ii)-RNG4; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),240))/16;
            if(data(ii) >= RNG3) data(ii) = data(ii)-RNG4; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+1),15));
            if(data(ii) >= RNG3) data(ii) = data(ii)-RNG4; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),240))/16;
            if(data(ii) >= RNG3) data(ii) = data(ii)-RNG4; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+2),15));
            if(data(ii) >= RNG3) data(ii) = data(ii)-RNG4; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+3),240))/16;
            if(data(ii) >= RNG3) data(ii) = data(ii)-RNG4; end
            ii = ii+1;data(ii)=double(bitand(buffer(i_d+3),15));
            if(data(ii) >= RNG3) data(ii) = data(ii)-RNG4; end
            i_d = i_d + 4;
        end
     end // of select mask
    end // of for nib
  end // of for i_f
      
  data(1) = sample_0;
  data = cumsum(data);
  sample_d = data(size(data,1));
  if(sample_d ~= sample_n)
    MS_file_info.err_flg = bitor(MS_file_info.err_flg,RNG5);
  end
  if(size(data,1) ~= MS_file_info.hdr(i_b,2))
    MS_file_info.err_flg = bitor(MS_file_info.err_flg,RNG6);
  end
//  mprintf('%d %d %d %d %d\n', i_b, sample_0, sample_n, data(1), data(size(data,1)));
  return;
endfunction


//----------------------------------
function [data] = Decode_DataS1(i_b)
  global MS_file_info;
  global buffer;
  data = [];
  ii = 0;
  mask = zeros(1,15);
  i_s = (i_b-1)*MS_file_info.rec_len+1;
  i_e = i_s + MS_file_info.rec_len - 1;
  if(MS_file_info.swap_flg)
    data_off = i_s + double(256*buffer(i_s+44)+buffer(i_s+45));
    sample_0 = ((iconvert(buffer(data_off+4),14)*256+iconvert(buffer(data_off+5),14))*256+iconvert(buffer(data_off+6),14))*256+iconvert(buffer(data_off+7),14);
    sample_n = ((iconvert(buffer(data_off+8),14)*256+iconvert(buffer(data_off+9),14))*256+iconvert(buffer(data_off+10),14))*256+iconvert(buffer(data_off+11),14);
  else
    data_off = i_s + double(256*buffer(i_s+45)+buffer(i_s+44));
    sample_0 = ((iconvert(buffer(data_off+7),14)*256+iconvert(buffer(data_off+6),14))*256+iconvert(buffer(data_off+5),14))*256+iconvert(buffer(data_off+4),14);
    sample_n = ((iconvert(buffer(data_off+11),14)*256+iconvert(buffer(data_off+10),14))*256+iconvert(buffer(data_off+9),14))*256+iconvert(buffer(data_off+8),14);
  end
  if(sample_0 >= MAXLONG) sample_0 = double(sample_0) - RNGLONG; end
  if(sample_n >= MAXLONG) sample_n = double(sample_n) - RNGLONG; end
  for i_f=data_off:LEN_FRAME:i_e
    if(MS_file_info.swap_flg)
      EC = ((iconvert(buffer(i_f),14)*256+iconvert(buffer(i_f+1),14))*256+iconvert(buffer(i_f+2),14))*256+iconvert(buffer(i_f+3),14);
    else
      EC = ((iconvert(buffer(i_f+3),14)*256+iconvert(buffer(i_f+2),14))*256+iconvert(buffer(i_f+1),14))*256+iconvert(buffer(i_f),14);
    end
//    mprintf('%d %d %d\n',sample_0, sample_n, EC);
    for nib=1:15
      mask(nib) = modulo(EC,4);
      EC = int(EC/4);
//     mprintf('%d ',mask(nib));
    end
    i_d = i_f+4;
    for nib=15:-1:1
      select mask(nib)
      case 0
        i_d=i_d+4;
        continue;
      case 1
        ii=ii+1;data(ii)=double(iconvert(buffer(i_d),1));i_d=i_d+1;
        ii=ii+1;data(ii)=double(iconvert(buffer(i_d),1));i_d=i_d+1;
        ii=ii+1;data(ii)=double(iconvert(buffer(i_d),1));i_d=i_d+1;
        ii=ii+1;data(ii)=double(iconvert(buffer(i_d),1));i_d=i_d+1;
      case 2
        if(MS_file_info.swap_flg)
          ii=ii+1;data(ii)=double(iconvert(buffer(i_d),11))*256+double(iconvert(buffer(i_d+1),11));i_d=i_d+2;
          if(data(ii) >= MAXSHORT) data(ii) = data(ii) - RNGSHORT; end
          ii=ii+1;data(ii)=double(iconvert(buffer(i_d),11))*256+double(iconvert(buffer(i_d+1),11));i_d=i_d+2;
          if(data(ii) >= MAXSHORT) data(ii) = data(ii) - RNGSHORT; end
        else
          ii=ii+1;data(ii)=double(iconvert(buffer(i_d+1),11))*256+double(iconvert(buffer(i_d),11));i_d=i_d+2;
          if(data(ii) >= MAXSHORT) data(ii) = data(ii) - RNGSHORT; end
          ii=ii+1;data(ii)=double(iconvert(buffer(i_d+1),11))*256+double(iconvert(buffer(i_d),11));i_d=i_d+2;
          if(data(ii) >= MAXSHORT) data(ii) = data(ii) - RNGSHORT; end
        end
      case 3
        if(MS_file_info.swap_flg)
          ii=ii+1;data(ii)=((double(iconvert(buffer(i_d),11))*256+double(iconvert(buffer(i_d+1),11)))*256+double(iconvert(buffer(i_d+2),11)))*256+double(iconvert(buffer(i_d+3),11));i_d=i_d+4;
          if(data(ii) >= MAXLONG) data(ii) = data(ii) - RNGLONG; end
        else
          ii=ii+1;data(ii)=((double(iconvert(buffer(i_d+3),11))*256+double(iconvert(buffer(i_d+2),11)))*256+double(iconvert(buffer(i_d+1),11)))*256+double(iconvert(buffer(i_d),11));i_d=i_d+4;
          if(data(ii) >= MAXLONG) data(ii) = data(ii) - RNGLONG; end
        end
      end        
    end
  end
  data(1) = sample_0;
  data = cumsum(data);
  sample_d = data(size(data,1));
  if(sample_d ~= sample_n)
    MS_file_info.err_flg = bitor(MS_file_info.err_flg,RNG5);
  end
  if(size(data,1) ~= MS_file_info.hdr(i_b,2))
    MS_file_info.err_flg = bitor(MS_file_info.err_flg,RNG6);
  end
//  mprintf('%d %d %d %d %d\n', i_b, sample_0, sample_n, data(1), data(size(data,1)));
  return;
endfunction


cd '/home/br/Scilab.sce/MSEED_UTILS';
tic();
[info,x] = RMSeed('in');
clf(0);
toc()
plot(x);
info


